#include "imgui/imgui.h"
#include "imgui/imgui_impl_opengl3.h"
#include "imgui/imgui_impl_sdl.h"
#include "imgui/imgui_impl_sdlrenderer.h"
#include "imgui/imgui_stdlib.h"
#include "imgui/implot.h"
#include <SDL2/SDL.h>
#include <SDL2/SDL_opengl.h>
#include <cmath>
#include <cstdio>
#include <cstring>
#include <iterator>

#include <algorithm>
#include <array>
#include <fstream>
#include <iostream>
#include <list>
#include <random>
#include <sstream>
#include <string>
#include <type_traits>
#include <vector>

//import im_NoiseFunctions/im_NoiseFunctions;
//import math;

void ShowTextCharactersDemo();

class ProductionTile {
public:
    ProductionTile();

    double production;
    double modifier;
};

class BaseTile {
public:
    BaseTile();

    int id;
    std::vector<int> tile_hash{0, 0, 0, 0};
    std::list<ProductionTile> production_tiles;
};

class TacticalMap {
public:
    TacticalMap();

    int map_extents;

};

class DataWindow {
public:
    DataWindow();

    void Render();

    void GenerateWindowTitle();

    int id;
    std::string message;
    bool isVisible;
    bool kill;
    char *window_title;

private:
};

class app {
public:
    app();

    int OnExecute();

    bool OnInit();

    void OnEvent(SDL_Event *event);

    void OnLoop();

    void OnRender();

    void OnExit();

    std::list<DataWindow> dataWindowList;
private:
    // SDL2 Stuff.
    bool isRunning;
    SDL_Window *window;
    SDL_GLContext gl_context;
    SDL_Renderer *renderer;
    ImGuiIO *io2;
};

app::app() {
    dataWindowList = {};
    isRunning = true;
    window = nullptr;
    gl_context = nullptr;
}

double gnoise(double noise) {
    std::random_device randomDevice{};
    std::mt19937 generator{randomDevice()};
    std::normal_distribution<double> distribution{0, noise};
    return distribution(generator);
}

double enoise(double noise) {
    std::random_device randomDevice{};
    std::mt19937 generator{randomDevice()};
    std::uniform_real_distribution<double> distribution(1, noise);
    return distribution(generator);
}

int enoise(int noise) {
    std::random_device randomDevice{};
    std::mt19937 generator{randomDevice()};
    std::uniform_int_distribution<int> distribution(1, noise);
    return distribution(generator);
}

int Roll20(int mod = 0) {
    return enoise(20) + mod;
}

std::string LoadTextFromFile() {
    std::ifstream t("file.txt");
    std::stringstream buffer;
    buffer << t.rdbuf();
    return buffer.str();
}

void ShiftArrayBack(double (&original)[50]) {
    int original_Size = sizeof(original) / sizeof(original[0]);

    for (int i = 0; i < original_Size; i++) {
        original[i] = original[i + 1];
    }
    if (original[48] > 10.0 || original[48] < -10.0) {
        original[48] = 0.0;
    }
    original[49] = original[48] + gnoise(1.0);
}

void AddValueToLIFOVector(double value, std::vector<double> &LIFO_vector) {
    uint size = LIFO_vector.size();
    for (int i = 0; i < size; i++) {
        LIFO_vector[i] = LIFO_vector[i + 1];
    }
    LIFO_vector[size-1] = value;
}

void DrawResourceProductionDemo() {
    static BaseTile newBase;

    ImGui::Begin(" Resource Production Demo");

    ImGui::Text(" Base Tile: %i", newBase.id);
    static double resources = 0;
    static double production = 0;
    static double modifier = 0;
    static double last_production = 0;

    int i = 0;
    for (ProductionTile pt: newBase.production_tiles) {
        ImGui::Text(" Prod %i: %0.f (mod %0.f)", i, pt.production, pt.modifier);
        i++;
    }

    ImGui::Separator();
    ImGui::Text("Time: %0.1f", ImGui::GetTime());
    // ImGui::Text("Random: %0.1f", gnoise(20.0));
    static double refresh_time = 0.0;

    static std::vector<double> resource_vec(50);

    static int refresh_rate = 1.0;
    ImGui::Text("Refresh Rate (S/s)");
    ImGui::SameLine();
    ImGui::SliderInt("", &refresh_rate, 0.0, 60.0, "%.3f", 0);

    if (refresh_time == 0.0)
        refresh_time = ImGui::GetTime();
    while (refresh_time < ImGui::GetTime() && refresh_rate != 0) {
        for (ProductionTile pt: newBase.production_tiles) {
            last_production = pt.production + gnoise(pt.modifier);
            resources += last_production;
        }
        AddValueToLIFOVector(resources, resource_vec);
        refresh_time += 1.0 / (double) refresh_rate;
    }
    double* r = &resource_vec[0];
    if (ImPlot::BeginPlot(" Production")) {
        ImPlot::PlotBars("Resource Gathering", r, resource_vec.size());
        ImPlot::EndPlot();
    }

    ImGui::Text(" Total Resources: %g (+%g)", resources, last_production);
    ImGui::Text(" Total Production: %g (max %g)", production, modifier + production);
    static std::string message = "";
    static double base = 0;
    static double over = 0;
    static double base_unit = 300.0;
    static double base_tile = 4000.0;
    static double over_unit = 0;
    static double over_tile = 0;
    if (ImGui::Button(" Buy New Unit") && resources > base_unit){
        message = "new unit";
        base = base_unit;
        over = std::abs(gnoise(100.0));
        resources -= base_unit + over;
    }
    if (ImGui::Button(" Build New Production Tile") && resources > base_tile){
        message = "new production tile";
        base = base_tile;
        over = std::abs(gnoise(2000.0));
        resources -= base_tile + over;
    }
    ImGui::Text("Purchased %s\n for %g with over-budget of %g",message.c_str(),base,over);
    ImGui::End();
}

void DrawDemoNoiseData() {
    ImGui::Text("Time: %0.1f", ImGui::GetTime());
    // ImGui::Text("Random: %0.1f", gnoise(20.0));
    static double refresh_time = 0.0;
    static double gnoise_arr[50] = {};
    static std::vector<double> noise_vec(50);
    std::vector<double> index_vec(50);
    std::iota(index_vec.begin(),index_vec.end(),0);


    static int refresh_rate = 1.0;
    ImGui::Text("Refresh Rate (S/s)");
    ImGui::SliderInt("##1", &refresh_rate, 0.0, 60.0, "%.3f", 0);
    double average = 0.0f;
    if (refresh_time == 0.0)
        refresh_time = ImGui::GetTime();
    while (refresh_time < ImGui::GetTime() && refresh_rate != 0) {
        AddValueToLIFOVector(gnoise(10), noise_vec);
        ShiftArrayBack(gnoise_arr);
        for (double n: gnoise_arr) {
            average += n;
        }
        average /= (double) IM_ARRAYSIZE(gnoise_arr);
        refresh_time += 1.0 / (double) refresh_rate;
    }

    double* x = &index_vec[0];
    double* y = &noise_vec[0];

    if (ImPlot::BeginPlot("Noise Stream")){
        //ImPlot::PlotBars("Noise Stream", gnoise_arr, IM_ARRAYSIZE(gnoise_arr));
        ImPlot::PlotBars("Noise Stream", y, IM_ARRAYSIZE(y));
        ImPlot::PlotLine("My Line Plot", y, x, IM_ARRAYSIZE(x));
        ImPlot::EndPlot();
    }


    ImGui::Text("IOTA at 34? : %g", index_vec[34]);
    ImGui::Text("Noise at 34? : %g", noise_vec[34]);
    int p=0;
    for (double ind : index_vec){
        ImGui::Text("iota %g : noise %g", x[p],y[p]);
        p++;
    }
}

class Wave {
public:
    Wave();

    int id;
    std::string name;
    int dim;
    double start{};
    double delta{};
    std::vector<double> data;

private:
    int checksum{};
};

Wave::Wave() {
    id = 0;
    name = "";
    dim = 1;
    data = {};
}

class Variable {
public:
    Variable();

    int id;
    std::string name;
    float value;
    enum type {
    };

private:
    int checksum{};
};

Variable::Variable() {
    id = 0;
    name = "";
    value = 0.0;
}

class Unit {
public:
    Unit();

    int id;
    std::vector<int> initial_traits{0, 0, 0, 0};
    std::vector<int> traits{0, 0, 0, 0};
    std::vector<int> effects{0, 0, 0, 0};

    void SetStat(int trait_index, int amount);

    int GetDistanceFromStartingTraits();

    void UpdateUnitEffectsFromTile();

    std::string map_identifier;
    std::string name;
    int tile_pos_x;
    int tile_pos_y;
    std::vector<int> tile_hash{5, 12, 1, 9};
    int max_stat_shift;
    int health;
    int organization;
};

class Battle {
public:
    Battle(Unit &enemyUnit, Unit &ownUnit);

    Unit &atk_unit;
    Unit &def_unit;

    void RunBattleTurn();

    std::vector<int> tile_hash{4, 13, 0, 8};
};

Battle::Battle(Unit &enemyUnit, Unit &ownUnit) : atk_unit(ownUnit), def_unit(enemyUnit) {

}

void Battle::RunBattleTurn() {
    int own_atk = Roll20(atk_unit.effects[0]);
    int own_def = Roll20(atk_unit.effects[1]);
    int own_int = Roll20(atk_unit.effects[2]);
    int own_poi = Roll20(atk_unit.effects[3]);

    int ene_atk = Roll20(def_unit.effects[0]);
    int ene_def = Roll20(def_unit.effects[1]);
    int ene_int = Roll20(def_unit.effects[2]);
    int ene_poi = Roll20(def_unit.effects[3]);
/*
    std::cout << own_atk << "\n";
    std::cout << own_def << "\n";
    std::cout << own_int << "\n";
    std::cout << own_poi << "\n";

    std::cout << ene_atk << "\n";
    std::cout << ene_def << "\n";
    std::cout << ene_int << "\n";
    std::cout << ene_poi << "\n";
*/
    int damage_to_own = (own_atk - ene_def) < 0 ? 0 : (own_atk - ene_def);
    int damage_to_ene = (ene_atk - own_def) < 0 ? 0 : (ene_atk - own_def);

    int disorg_to_own = (own_int - ene_poi) < 0 ? 0 : (own_int - ene_poi);
    int disorg_to_ene = (ene_int - own_poi) < 0 ? 0 : (ene_int - own_poi);

    std::cout << "Damage to Own " << damage_to_own << "\n";
    std::cout << "Damage to Ene " << damage_to_ene << "\n";
    std::cout << "Disorg to Own " << disorg_to_own << "\n";
    std::cout << "Disorg to Ene " << disorg_to_ene << "\n";

    atk_unit.health -= damage_to_own;
    def_unit.health -= damage_to_ene;
    atk_unit.organization -= disorg_to_own;
    def_unit.organization -= disorg_to_ene;

    std::cout << "Own Health: " << atk_unit.health << "\n";
    std::cout << "Own Organization: " << atk_unit.organization << "\n";
    std::cout << "Ene Health: " << def_unit.health << "\n";
    std::cout << "Ene Organization: " << def_unit.organization << "\n\n";
}

void DrawSingleUnitStatsWindow(Unit &);

class DataBrowser {
public:
    DataBrowser();

    std::list<Wave> waves;
    std::list<Variable> variables;
    std::list<Wave>::iterator wave_iter;
    std::list<Wave>::iterator iter_variable;

    void KillData();
};

DataBrowser::DataBrowser() {
    waves = {};
    variables = {};
    wave_iter = {};
    iter_variable = {};
}

void DataBrowser::KillData() { return; }


DataWindow::DataWindow() {
    isVisible = false;
    kill = false;
    message = "empty";
    id = 0;
    window_title = 0;
}

void DataWindow::Render() {
    if (isVisible == false) {
        return;
    }
    ImGui::Begin(window_title);
    ImGui::Text("%s", message.c_str());
    if (ImGui::Button("Ok.")) {
        std::cout << "button: isRunning false\n";
        isVisible = false;
    }
    ImGui::Checkbox("isRunning", &isVisible);
    // ImGui::SliderFloat("", &randomTag, -3.0, 10.0, "%.3f");
    ImGui::End();
}

void DataWindow::GenerateWindowTitle() {
    std::string standardHeader = "Window ";
    std::string imguiSpecialWindowCode = "##";
    std::string idNumberString = std::to_string(id);
    std::string combo =
            standardHeader + idNumberString + imguiSpecialWindowCode + idNumberString;
    char *newWindowTitleChar = new char[combo.size() + 1];
    std::copy(combo.begin(), combo.end(), newWindowTitleChar);
    newWindowTitleChar[combo.size()] = '\0';
    window_title = newWindowTitleChar;
    std::cout << "generate window: isRunning true\n";
    isVisible = true;
}

void DrawWindowNewMessage(std::string message) {
    static int counter = 0;
    ImGui::Begin("Made it##");
    ImGui::Text("%s", message.c_str());
    ImGui::Text("%d", counter);
    ImGui::End();
    counter++;
}

void SpawnNewWindowWithMessage(char *message, std::list<DataWindow> &list) {
    std::string editableMessage(message);
    if (editableMessage.size() == 0) {
        return;
    }
    //std::list<DataWindow> here = dataWindowList;
    std::cout << "called new win message\n";
    std::cout << "start list size: " << list.size() << "\n";
    static int counter = 1;
    DataWindow nWin;
    nWin.message = editableMessage;
    nWin.id = counter;
    nWin.GenerateWindowTitle();
    list.push_back(nWin);
    counter++;
    std::cout << "End Size: " << list.size() << "\n";
}

void SpawnNewWindowMessage(std::string message, std::list<DataWindow> &list) {
    std::string editableMessage = message;
    if (editableMessage.size() == 0) {
        return;
    }
    std::cout << "called new win message\n";
    std::cout << "start list size: " << list.size() << "\n";
    static int counter = 1;
    DataWindow nWin;
    nWin.message = editableMessage;
    nWin.id = counter;
    nWin.GenerateWindowTitle();
    list.push_back(nWin);
    counter++;
    std::cout << "End Size: " << list.size() << "\n";
}

void FlagWindowForKillById(int id, std::list<DataWindow> &list) {
    auto i = list.begin();
    for (DataWindow &nWin: list) {
        if (nWin.id == id) {
            nWin.kill = true;
            return;
        }
        i++;
    }
    std::cout << "> [ERROR] No Window Found"
              << "\n";
}

void ProcessKillFlags(std::list<DataWindow> &list) {
    auto i = list.begin();
    for (DataWindow &nWin: list) {
        if (nWin.kill) {
            list.erase(i);
            return;
        }
        i++;
    }
}

void KillWave(int id, DataBrowser &dataBrowser) {
    // ... both algos work without breaking!
    /* std::list<Wave>::iterator iter =
        std::remove_if(DataBrowser.waves.begin(), DataBrowser.waves.end(),
                       [id](Wave n) { return n.id == id; });

    DataBrowser.waves.erase(iter, DataBrowser.waves.end()); */

    // ... this loop works without breaking.
    // ... uses an internal iterator inside of the DataBrowser object
    // ... compare the current iterator with the ImGui widget identifier
    for (auto it = dataBrowser.waves.begin(); it != dataBrowser.waves.end();) {
        if (it->id == id) {
            it = dataBrowser.waves.erase(it);
            dataBrowser.wave_iter = it;
        } else {
            ++it;
        }
    }
}

// Make button to kill Wave
void DrawKillWaveButton(int id, DataBrowser &dataBrowser) {
    std::string button_message = "Kill";
    std::string id_fill = "##" + std::to_string(id);
    std::string button_id = button_message + id_fill;
    if (ImGui::SmallButton(button_id.c_str())) {
        KillWave(id, dataBrowser);
    }
}

// Make button to kill window
void FlagWindowForKillButton(int id, std::list<DataWindow> &list) {
    std::string button_message = "Kill";
    std::string id_fill = "##" + std::to_string(id);
    std::string button_id = button_message + id_fill + std::to_string(id);
    if (ImGui::Button(button_id.c_str())) {
        FlagWindowForKillById(id, list);
    }
}

void DrawAllListMembers(std::list<DataWindow> &list) {
    ImGui::Text("Currently Loaded Windows");
    ImGui::SameLine();
    ImGui::Text("%zu", list.size());
    for (DataWindow &dWin: list) {
        FlagWindowForKillButton(dWin.id, list);
        ImGui::SameLine();
        std::string id_fill = "##" + std::to_string(dWin.id);
        ImGui::Checkbox(id_fill.c_str(), &dWin.isVisible);
        ImGui::SameLine();
        ImGui::Text("%s", dWin.window_title);
        ImGui::SameLine();
        ImGui::Text(": %s", dWin.message.c_str());
    }
}

void DrawLoadedWaves(DataBrowser &dataBrowser) {
    ImGui::SetNextTreeNodeOpen(true);
    if (ImGui::TreeNode("Wave Browser")) {
        for (dataBrowser.wave_iter = dataBrowser.waves.begin();
             dataBrowser.wave_iter != dataBrowser.waves.end();
             ++dataBrowser.wave_iter) {
            DrawKillWaveButton(dataBrowser.wave_iter->id, dataBrowser);
            ImGui::SameLine();
            ImGui::Text("Wave: %s", dataBrowser.wave_iter->name.c_str());
            ImGui::SameLine();
            float *dataPreviewArray = reinterpret_cast<float *>(&dataBrowser.wave_iter->data[0]);
            int dataLength = dataBrowser.wave_iter->data.size();
            ImGui::PlotLines("", dataPreviewArray, dataLength);
        }
        ImGui::TreePop();
    }
}

Wave MakeNewWave(std::string wave_name, int size) {
    std::cout << "size" << size << "\n";
    static int counter = 0;
    Wave wave;
    wave.id = counter;
    wave.name = wave_name;
    std::vector<double> fakeNoiseData(size, 0.0);
    for (int i = 0; i < size; i++) {
        fakeNoiseData[i] = gnoise(1);
    }
    wave.data = fakeNoiseData;
    counter++;
    return wave;
}

// Main code
void app::OnLoop() {

    static std::string textMsg = "hallo";
    static std::list<DataWindow> winList;
    static DataBrowser dataBrowser;

    ImGui::ShowDemoWindow();
    ImPlot::ShowDemoWindow();


    DrawResourceProductionDemo();
    ShowTextCharactersDemo();


    // ... Bil Window Be Low
    {
        ImGui::Begin("Data Browser");
        ImGui::Text("Wave Making Area");
        static int wave_count = 0;
        if (ImGui::Button("Make a new random Wave")) {
            std::string new_wave_name = "Wave " + std::to_string(wave_count);
            dataBrowser.waves.push_back(MakeNewWave(new_wave_name.c_str(), enoise(100)));
            ++wave_count;
        }
        ImGui::Text("Formatting errors: %s, %a", textMsg.c_str(), textMsg.c_str());

        static char new_window_message_buffer[32] = "";
        static const ImGuiInputTextFlags flags =
                ImGuiInputTextFlags_EnterReturnsTrue;
        if (ImGui::Button("Spawn Window")) {
            SpawnNewWindowWithMessage(new_window_message_buffer, winList);
            strcpy(new_window_message_buffer, "");
        }
        ImGui::SameLine();
        if (ImGui::InputTextWithHint(
                "", "enter window name...", new_window_message_buffer,
                IM_ARRAYSIZE(new_window_message_buffer), flags)) {
            SpawnNewWindowWithMessage(new_window_message_buffer, winList);
            strcpy(new_window_message_buffer, "");
        }

        if (ImGui::Button("Load Text From file")) {
            SpawnNewWindowMessage(LoadTextFromFile(), winList);
        }
        ImGui::Separator();
        static int killWindowIndex = 0;
        ImGui::SetNextItemWidth(100);
        ImGui::InputInt("Window Id", &killWindowIndex);
        ImGui::SameLine();
        if (ImGui::Button("Kill Window")) {
            FlagWindowForKillById(killWindowIndex, winList);
        }

        ImGui::SameLine();
        ImGui::Text("%zu Windows Exist", winList.size());

        if (ImGui::TreeNode("Loaded Windows")) {
            DrawAllListMembers(winList);
            ImGui::TreePop();
        }

        DrawDemoNoiseData();
        DrawLoadedWaves(dataBrowser);

        static Unit unit_1;
        unit_1.id = 1;
        static Unit unit_2;
        unit_2.id = 2;
        DrawSingleUnitStatsWindow(unit_1);
        DrawSingleUnitStatsWindow(unit_2);

        static Battle testBattle(unit_2, unit_1);
        if (ImGui::Button("Run One Battle...")) {
            testBattle.RunBattleTurn();
        }

        ImGui::End();
    }

    // ... Render all tracked window objects
    // ... things I've learned
    // ... references are a real thing to use and goddamn are they important
    // ... this loop is meant to take all the newWindow objects in
    // ... winList and run their render function
    for (DataWindow &nWin: winList) {
        nWin.Render();
    }
    ProcessKillFlags(winList);
}


bool app::OnInit() {
    if (SDL_Init(SDL_INIT_EVERYTHING) < 0) {
        return false;
    }

    if (1) {
        auto window_flags =
                (SDL_WindowFlags) (SDL_WINDOW_OPENGL | SDL_WINDOW_RESIZABLE |
                                   SDL_WINDOW_ALLOW_HIGHDPI);
        window =
                SDL_CreateWindow("Welcome to the Terror Dome", SDL_WINDOWPOS_CENTERED,
                                 SDL_WINDOWPOS_CENTERED, 1100, 1300, window_flags);
        renderer = SDL_CreateRenderer(window, -1, SDL_RENDERER_PRESENTVSYNC | SDL_RENDERER_ACCELERATED);
        if (renderer == nullptr) {
            SDL_Log("Error creating SDL_Renderer!");
            return false;
        }

        IMGUI_CHECKVERSION();
        ImGui::CreateContext();
        ImPlot::CreateContext();
        ImGuiIO &io = ImGui::GetIO();
        (void) io;

        ImVector<ImWchar> ranges;
        ImWchar privater[] = {0xE000,0xF8FF}; // icon range for monofur
        ImFontGlyphRangesBuilder builder;
        builder.AddText("","");                        // Add a string (here "Hello world" contains 7 unique characters)
        builder.AddChar(0xF939);                               // Add a specific character
        builder.AddRanges(privater);
        builder.AddRanges(io.Fonts->GetGlyphRangesDefault()); // Add one of the default ranges
        builder.BuildRanges(&ranges);

        io.Fonts->AddFontFromFileTTF("monof.ttf", 20, nullptr,ranges.Data);
        io.Fonts->Build();

        ImGui::StyleColorsDark();

        ImGui_ImplSDL2_InitForSDLRenderer(window);
        ImGui_ImplSDLRenderer_Init(renderer);
    } else {
        // GL versions
        const char *glsl_version = "#version 150";
        SDL_GL_SetAttribute(SDL_GL_CONTEXT_FLAGS, 0);
        SDL_GL_SetAttribute(SDL_GL_CONTEXT_PROFILE_MASK, SDL_GL_CONTEXT_PROFILE_CORE);
        SDL_GL_SetAttribute(SDL_GL_CONTEXT_MAJOR_VERSION, 3);
        SDL_GL_SetAttribute(SDL_GL_CONTEXT_MINOR_VERSION, 2);
        // Create Window
        SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
        SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 24);
        SDL_GL_SetAttribute(SDL_GL_STENCIL_SIZE, 8);
        auto window_flags =
                (SDL_WindowFlags) (SDL_WINDOW_OPENGL | SDL_WINDOW_RESIZABLE |
                                   SDL_WINDOW_ALLOW_HIGHDPI);
        window =
                SDL_CreateWindow("Welcome to the Terror Dome", SDL_WINDOWPOS_CENTERED,
                                 SDL_WINDOWPOS_CENTERED, 1100, 1300, window_flags);
        gl_context = SDL_GL_CreateContext(window);
        SDL_GL_MakeCurrent(window, gl_context);
        if (window != nullptr) {
            std::cout << "\n> Initialized Window .....\n";
        } else {
            return false;
        }
        // ImGUI stuff??
        IMGUI_CHECKVERSION();
        ImGui::CreateContext();
        ImPlot::CreateContext();
        ImGuiIO &io = ImGui::GetIO();
        (void) io;
        ImGui::StyleColorsDark();
        //ImGui_ImplSDL2_InitForSDLRenderer(window);
        ImGui_ImplOpenGL3_Init(glsl_version);
    }
    return true;
}

int app::OnExecute() {
    SDL_Event event;

    if (!OnInit()) {
        return -1;
    }

    while (isRunning) {
        while (SDL_PollEvent(&event) != 0) {
            OnEvent(&event);
            ImGui_ImplSDL2_ProcessEvent(&event);
        }
        if (1) {
            ImGui_ImplSDLRenderer_NewFrame();
            ImGui_ImplSDL2_NewFrame(window);
        } else {
            ImGui_ImplOpenGL3_NewFrame();
            ImGui_ImplSDL2_NewFrame();
        }
        ImGui::NewFrame();
        OnLoop();
        OnRender();
    }
    return 0;
}

void app::OnEvent(SDL_Event *event) {
    if (event->type == SDL_QUIT) {
        isRunning = false;
    }
}

void app::OnRender() {
    ImGui::Render();
    if (1) {
        ImVec4 clear_color = ImVec4(0.45f, 0.55f, 0.60f, 1.00f);
        SDL_SetRenderDrawColor(renderer, (Uint8) (clear_color.x * 255), (Uint8) (clear_color.y * 255),
                               (Uint8) (clear_color.z * 255), (Uint8) (clear_color.w * 255));
        SDL_RenderClear(renderer);
        ImGui_ImplSDLRenderer_RenderDrawData(ImGui::GetDrawData());
        SDL_RenderPresent(renderer);
    } else {
        glViewport(0, 0, 1280, 720);
        glClearColor(0.45f, 0.55f, 0.60f, 1.0f);
        glClear(GL_COLOR_BUFFER_BIT);
        ImGui_ImplOpenGL3_RenderDrawData(ImGui::GetDrawData());
        SDL_GL_SwapWindow(window);
    }

}

void app::OnExit() {
    std::cout << "> Exiting...";
    //ImGui_ImplOpenGL3_Shutdown();
    ImGui_ImplSDLRenderer_Shutdown();
    ImGui_ImplSDL2_Shutdown();
    ImGui::DestroyContext();
    ImPlot::DestroyContext();

    //SDL_GL_DeleteContext(gl_context);
    SDL_DestroyRenderer(renderer);
    //gl_context = nullptr;
    renderer = nullptr;
    SDL_DestroyWindow(window);
    window = nullptr;
    SDL_Quit();
}

int main(int argc, char *argv[]) {
    app App;
    return App.OnExecute();
}


ProductionTile::ProductionTile() {
    production = 0.0;
    modifier = 0.0;
}

BaseTile::BaseTile() {
    id = 1;
    for (int n: tile_hash) {
        static int i = 0;
        tile_hash[i] = enoise(15);
        i++;
    }
    int numNewTiles = enoise(5);
    for (int i = 0; i < numNewTiles; i++) {
        ProductionTile newProdTile;
        newProdTile.production = 10.0 + enoise(30);
        newProdTile.modifier = 5.0 + enoise(20);
        production_tiles.push_back(newProdTile);
    }
}

TacticalMap::TacticalMap() {

}


Unit::Unit() {
    for (int n: traits) {
        static int i = 0;
        traits[i] = enoise(15);
        i++;
    }
    initial_traits = traits;
    max_stat_shift = 11;
    UpdateUnitEffectsFromTile();
    health = 100;
    organization = 100;
}

int Unit::GetDistanceFromStartingTraits() {
    int total_distance = 0;
    int i = 0;
    for (int n: traits) {
        total_distance += std::abs(initial_traits[i] - traits[i]);
        i++;
    }
    return total_distance;
}

void Unit::UpdateUnitEffectsFromTile() {
    int i = 0;
    for (int n: traits) {
        effects[i] = 15 - std::abs(tile_hash[i] - traits[i]);
        i++;
    }
}

void Unit::SetStat(int trait_index, int amount) {
    // ... check if the stats exceed bounds
    int o = traits[trait_index] + amount;
    if (o < 0 || o > 15) {
        std::cout << "Stat Bounds Reached" << "\n";
        return;
    }
    // ... all good? change the stat, register the change
    traits[trait_index] += amount;
    // ... compare moves to total stats shift allowed
    // ... ... reverse course if problems occur - skip update
    if (GetDistanceFromStartingTraits() > max_stat_shift) {
        traits[trait_index] -= amount;
        std::cout << "This Unit doesn't have the capacity to change any more" << "\n";
        return;
    }
    // ... update against current tile hash
    UpdateUnitEffectsFromTile();
}

std::string GetTraitNameByIndex(int index){
    switch (index){
        case 0:
            return  "";
        case 1:
            return  "";
        case 2:
            return  "";
        case 3:
            return  "";
        default:
            return "no trait";
    }
}


void DrawSingleUnitStatBlock(int trait_index, Unit &currentUnit) {
    //trait[statIndex];
    ImGui::PushID(trait_index);
    if (ImGui::SmallButton("-1")) {
        currentUnit.SetStat(trait_index, -1);
    }
    ImGui::SameLine();
    if (ImGui::SmallButton(("+1"))) {
        currentUnit.SetStat(trait_index, 1);
    }
    ImGui::SameLine();
    std::string trait_name = "new trait";
    ImGui::Text("%s: %02.1i (%02.1i) %+02.1i",
                GetTraitNameByIndex(trait_index).c_str(),
                currentUnit.traits[trait_index],
                currentUnit.initial_traits[trait_index],
                currentUnit.effects[trait_index]);
    ImGui::PopID();
}

void DrawSingleUnitStatsWindow(Unit &currentUnit) {

    std::string id_str = std::to_string(currentUnit.id);
    std::string id_fill = "##" + id_str;
    std::string begin_str = " Unit #" + id_str + id_fill;

    ImGui::Begin(begin_str.c_str());

    ImGui::Text("Unit %i info!", currentUnit.id);

    ImGui::Separator();
    ImGui::Text(": %i", currentUnit.health);
    ImGui::Text(": %i", currentUnit.organization);
    // TODO: roll into a loop for rendering (see Wave display above for help)
    // TODO: make enum for traits
    int i_index = 0;
    for (int n : currentUnit.traits){
        DrawSingleUnitStatBlock(i_index, currentUnit);
        ++i_index;
    }

    ImGui::Separator();
    ImGui::Text("Stat Changes Available: %i / %i", (11- currentUnit.GetDistanceFromStartingTraits()),
                currentUnit.max_stat_shift);


    ImGui::End();
}


void ShowTextCharactersDemo() {
    ImGui::Begin("Unique Characters Demo");

    ImGui::Text("Let's Go!");

    ImGui::End();
}